/*-
 * Copyright (c) 1999 James Howard and Dag-Erling Codan Smrgrav
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>

#ifndef lint
#if 0
__RCSID("$OpenBSD: file.c,v 1.9 2006/02/09 09:54:46 otto Exp $");
#else
__FBSDID("$FreeBSD$");
#endif
#endif /* not lint */

#include <sys/param.h>

#include <bzlib.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <zlib.h>

#include "grep.h"

static char	 fname[MAXPATHLEN];

#define FILE_STDIO	0
#define FILE_MMAP	1

struct file {
	int		 type;
	int		 noseek;
	FILE		*f;
	struct mmfile	*mmf;
};

struct file *
grep_stdin_open(void)
{
	struct file	*f;

	snprintf(fname, sizeof fname, getstr(1));

	f = grep_malloc(sizeof *f);

	f->type = FILE_STDIO;
	f->noseek = isatty(FILE_STDIO);
	if ((f->f = fdopen(FILE_STDIO, "r")) != NULL)
		return (f);

	free(f);
	return (NULL);
}

struct file *
grep_open(char *path)
{
	struct file	*f;
	char		*templ;
	int		 tempfd;

	snprintf(fname, sizeof fname, "%s", path);

	f = grep_malloc(sizeof *f);
	f->noseek = 0;

	if (Zflag || Jflag) {
		templ = grep_malloc(sizeof(char) * 15);
		strlcpy(templ, "/tmp/grep.XXXXXXXX", 14);
		if ((tempfd = mkstemp(templ)) == -1)
			err(2, NULL);
		free(templ);
		if (Zflag) {
			gzFile	*gzf;
			char	 buf[BUFSIZ];
			int	 i;

			if ((gzf = gzopen(fname, "r")) == NULL)
				err(2, NULL);
			while ((i = gzread(gzf, buf, BUFSIZ)) > 0) {
				write(tempfd, buf, BUFSIZ);
			}
			gzclose(gzf);
			lseek(tempfd, 0L, SEEK_SET);
			f->type = FILE_STDIO;
			if ((f->f = fdopen(tempfd, "r")) != NULL)
				return (f);
			else
				return (NULL);
		} else {
			BZFILE	*bzf;
			char	 buf[BUFSIZ];
			int	 bzerror;
			FILE	*file;

			if ((file = fopen(fname, "r")) == NULL)
				err(2, NULL);
			if ((bzf = BZ2_bzReadOpen(&bzerror, file, 0, 0, NULL, 0)) == NULL)
				err(2, NULL);
			do {
				BZ2_bzRead(&bzerror, bzf, buf, BUFSIZ);
				write(tempfd, buf, BUFSIZ);
			} while (bzerror == BZ_OK);
			BZ2_bzReadClose(&bzerror, bzf);
			fclose(file);
			lseek(tempfd, 0L, SEEK_SET);
			f->type = FILE_STDIO;
			if ((f->f = fdopen(tempfd, "r")) != NULL)
				return (f);
			else
				return (NULL);
		}
	}

	/* try mmap first; if it fails, try stdio */
	if ((f->mmf = mmopen(fname, "r")) != NULL) {
		f->type = FILE_MMAP;
		return (f);
	}
	f->type = FILE_STDIO;
	if ((f->f = fopen(path, "r")) != NULL)
		return (f);

	free(f);
	return (NULL);
}

int
grep_bin_file(struct file *f)
{
	if (f->noseek)
		return (0);

	switch (f->type) {
	case FILE_STDIO:
		return (bin_file(f->f));
	case FILE_MMAP:
		return (mmbin_file(f->mmf));
	default:
		/* NOTREACHED */
		errx(2, getstr(3));
	}
}

char *
grep_fgetln(struct file *f, size_t *l)
{
	switch (f->type) {
	case FILE_STDIO:
		return (fgetln(f->f, l));
	case FILE_MMAP:
		return (mmfgetln(f->mmf, l));
	default:
		/* NOTREACHED */
		errx(2, getstr(3));
	}
}

void
grep_close(struct file *f)
{
	switch (f->type) {
	case FILE_STDIO:
		fclose(f->f);
		break;
	case FILE_MMAP:
		mmclose(f->mmf);
		break;
	default:
		/* NOTREACHED */
		errx(2, getstr(3));
	}
	free(f);
}
