/*-
 * Copyright (c) 1999 James Howard and Dag-Erling Codan Smrgrav
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>

#ifndef lint
#if 0
__RCSID("$OpenBSD: grep.c,v 1.39 2007/09/02 15:19:32 deraadt Exp $");
#else
__FBSDID("$FreeBSD$");
#endif
#endif /* not lint */

#include <sys/limits.h>
#include <sys/queue.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <getopt.h>
#include <locale.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "grep.h"

#ifndef WITHOUT_NLS
#include <nl_types.h>
nl_catd	 catalog;
#endif

char	*errstr[] = {
	"",
/* 1*/	"(standard input)",
/* 2*/	"(fd %d)",
/* 3*/	"invalid file type",
/* 4*/	"usage: %s [-abcDEFGHhIiJLlmnOoPqRSsUVvwxZ] [-A num] [-B num] [-C[num]]\n",
/* 5*/	"\t[-e pattern] [-f file] [--binary-files=value] [--color=when]\n",
/* 6*/	"\t[--context[=num]] [--directories=action] [--label] [--line-buffered]\n",
/* 7*/	"\t[--null] [pattern] [file ...]\n",
/* 8*/	"parentheses not balanced",
/* 9*/	"context out of range",
/*10*/	"FreeBSD grep 2.5.1\n",
/*11*/	"unknown --binary-files option",
/*12*/	"Binary file %s matches\n",
/*12*/	"value out of range",
/*13*/	"unknown -d or --directory option",
/*14*/	"unknown --color option"
};

/* Flags passed to regcomp() and regexec() */
int		 cflags;
int		 eflags = REG_STARTEND;

int		 matchall;	/* shortcut */
int		 patterns, pattern_sz;
char		**pattern;
regex_t		*r_pattern;
struct fastgrep	*fg_pattern;

/* For regex errors  */
char	 re_error[RE_ERROR_BUF + 1];

/* Command-line flags */
unsigned long long Aflag;	/* -A x: print x lines trailing each match */
unsigned long long Bflag;	/* -B x: print x lines leading each match */
int	 Eflag;		/* -E: interpret pattern as extended regexp */
int	 Fflag;		/* -F: interpret pattern as list of fixed strings */
int	 Gflag;		/* -G: interpret pattern as basic regexp */
int	 Hflag;		/* -H: always print file name */
int	 Jflag;		/* -J: grep in bzipped file */
int	 Lflag;		/* -L: only show names of files with no matches */
int	 Zflag;		/* -Z: grep in gzipped file */
int	 bflag;		/* -b: show block numbers for each match */
int	 cflag;		/* -c: only show a count of matching lines */
int	 hflag;		/* -h: don't print filename headers */
int	 iflag;		/* -i: ignore case */
int	 lflag;		/* -l: only show names of files with matches */
int	 mflag;		/* -m x: stop reading the files after x matches */
int	 nflag;		/* -n: show line numbers in front of matching lines */
int	 oflag;		/* -o: print only matching part */
int	 qflag;		/* -q: quiet mode (don't output anything) */
int	 sflag;		/* -s: silent mode (ignore errors) */
int	 vflag;		/* -v: only show non-matching lines */
int	 wflag;		/* -w: pattern must start and end on word boundaries */
int	 xflag;		/* -x: pattern must match entire line */
int	 lbflag;	/* --line-buffered */
int	 nullflag;	/* --null */
char	*label;		/* --label */
char	*color;		/* --color */
unsigned long long mcount;	/* count for -m */

int	 binbehave = BINFILE_BIN;
int	 devbehave = DEV_GREP;
int	 dirbehave = DIR_GREP;
int	 linkbehave = LINK_GREP;

enum {
	BIN_OPT = CHAR_MAX + 1,
	COLOR_OPT,
	HELP_OPT,
	MMAP_OPT,
	LINEBUF_OPT,
	LABEL_OPT,
	NULL_OPT
};

/* Housekeeping */
int	 first = 1;	/* flag whether we are processing the first match */
int	 prev;		/* flag whether or not the previous line matched */
int	 tail;		/* lines left to print */
int	 notfound;	/* file not found */

struct patfile {
	const char		*pf_file;
	SLIST_ENTRY(patfile)	 pf_next;
};
SLIST_HEAD(, patfile)		 patfilelh;

extern char	*__progname;

static void
usage(void)
{
	fprintf(stderr, getstr(4), __progname);
	fprintf(stderr, getstr(5));
	fprintf(stderr, getstr(5));
	fprintf(stderr, getstr(6));
	fprintf(stderr, getstr(7));
	exit(2);
}

static char	*optstr = "0123456789A:B:C:D:EFGHIJLOPSRUVZabcd:e:f:hilm:noqrsuvwxy";

struct option long_options[] =
{
	{"binary-files",	required_argument,	NULL, BIN_OPT},
	{"help",		no_argument,		NULL, HELP_OPT},
	{"mmap",		no_argument,		NULL, MMAP_OPT},
	{"line-buffered",	no_argument,		NULL, LINEBUF_OPT},
	{"label",		required_argument,	NULL, LABEL_OPT},
	{"null",		no_argument,		NULL, NULL_OPT},
	{"color",		optional_argument,	NULL, COLOR_OPT},
	{"colour",		optional_argument,	NULL, COLOR_OPT},
	{"after-context",	required_argument,	NULL, 'A'},
	{"text",		no_argument,		NULL, 'a'},
	{"before-context",	required_argument,	NULL, 'B'},
	{"byte-offset",		no_argument,		NULL, 'b'},
	{"context",		optional_argument,	NULL, 'C'},
	{"count",		no_argument,		NULL, 'c'},
	{"devices",		required_argument,	NULL, 'D'},
        {"directories",		required_argument,	NULL, 'd'},
	{"extended-regexp",	no_argument,		NULL, 'E'},
	{"regexp",		required_argument,	NULL, 'e'},
	{"fixed-strings",	no_argument,		NULL, 'F'},
	{"file",		required_argument,	NULL, 'f'},
	{"basic-regexp",	no_argument,		NULL, 'G'},
	{"no-filename",		no_argument,		NULL, 'h'},
	{"with-filename",	no_argument,		NULL, 'H'},
	{"ignore-case",		no_argument,		NULL, 'i'},
	{"bz2decompress",	no_argument,		NULL, 'J'},
	{"files-with-matches",	no_argument,		NULL, 'l'},
	{"files-without-match", no_argument,            NULL, 'L'},
	{"max-count",		required_argument,	NULL, 'm'},
	{"line-number",		no_argument,		NULL, 'n'},
	{"only-matching",	no_argument,		NULL, 'o'},
	{"quiet",		no_argument,		NULL, 'q'},
	{"silent",		no_argument,		NULL, 'q'},
	{"recursive",		no_argument,		NULL, 'r'},
	{"no-messages",		no_argument,		NULL, 's'},
	{"binary",		no_argument,		NULL, 'U'},
	{"unix-byte-offsets",	no_argument,		NULL, 'u'},
	{"invert-match",	no_argument,		NULL, 'v'},
	{"version",		no_argument,		NULL, 'V'},
	{"word-regexp",		no_argument,		NULL, 'w'},
	{"line-regexp",		no_argument,		NULL, 'x'},
	{"decompress",          no_argument,            NULL, 'Z'},
	{NULL,			no_argument,		NULL, 0}
};


static void
add_pattern(char *pat, size_t len)
{
	char	*ptr, *st;
	int	 lbr, rbr;

/* Workaround for our libc-regex library to match GNU behaviour.
   Our library rejects '|' with empty subexpressions.  Just cut out
   those parts, e.g. "(|a|b||c|)" will be "(a|b|c)" */
	st = pat;
	while (((ptr = strstr(st, "|)"))) != NULL) {
		if (strstr(st, "\\|)") != (ptr - 1))
			strlcpy(ptr, &(ptr[1]), strlen(ptr));
		st = ptr;
	}
	st = pat;
	while ((ptr = strstr(st, "(|")) != NULL) {
		if ((strstr(st, "(\\|") != ptr) && (strstr(st, "\\(|") != (ptr - 1)))
			strlcpy(&(ptr[1]), &(ptr[2]), strlen(ptr));
		st = ptr + 1;
	}
	st = pat;
	while ((ptr = strstr(st, "||")) != NULL) {
		if (strstr(st, "\\||") != (ptr - 1))
			strlcpy(&(ptr[1]), &(ptr[2]), strlen(ptr));
		st = ptr + 1;
	}
	st = pat;
	while ((ptr = strstr(st, "*?")) != NULL) {
		if (strstr(st, "\\*?") != (ptr - 1))
			strlcpy(&(ptr[1]), &(ptr[2]), strlen(ptr));
		st = ptr + 1;
	}
	st = pat;
	while ((ptr = strstr(st, "?*")) != NULL) {
		if (strstr(st, "\\?*") != (ptr - 1))
			strlcpy(&(ptr[1]), &(ptr[2]), strlen(ptr));
		st = ptr + 1;
	}
	st = pat;
	while ((ptr = strstr(st, "**")) != NULL) {
		if ((strstr(st, "\\**") != (ptr - 1)) && (strstr(st, "\\(**") != (ptr - 2)))
			strlcpy(&(ptr[1]), &(ptr[2]), strlen(ptr));
		st = ptr + 1;
	}
	st = pat;
	while ((ptr = strstr(st, "??")) != NULL) {
		if (strstr(st, "\\??") != (ptr - 1))
			strlcpy(&(ptr[1]), &(ptr[2]), strlen(ptr));
		st = ptr + 1;
	}
	st = pat;
	while ((ptr = strstr(st, "^*")) != NULL) {
		strlcpy(&(ptr[1]), &(ptr[2]), strlen(ptr));
		st = ptr + 1;
	}
	st = pat;
	while ((ptr = strstr(st, "(*")) != NULL) {
		if (strstr(st, "\\(*") != (ptr - 1))
			strlcpy(&(ptr[1]), &(ptr[2]), strlen(ptr));
		st = ptr + 1;
	}
	st = pat;
	if (Eflag) {
		st = pat;
		lbr = 0;
		while ((ptr = strstr(st, "(")) != NULL) {
			if (strstr(st, "\\(") != (ptr - 1))
				lbr++;
			st = ptr + 1;
		}
		st = pat;
		rbr = 0;
		while ((ptr = strstr(st, ")")) != NULL) {
			if (strstr(st, "\\)") != (ptr - 1))
				rbr++;
			st = ptr + 1;
		}
		if (lbr != rbr) {
			errx(2, getstr(8));
		}
	}

/* Work-around to allow * at the beginning of the regexp */
	while (pat[0] == '*')
		pat++;

	if (!xflag && (len == 0 || matchall)) {
		matchall = 1;
		return;
	}
	if (patterns == pattern_sz) {
		pattern_sz *= 2;
		pattern = grep_realloc(pattern, ++pattern_sz * sizeof(*pattern));
	}
	if (len > 0 && pat[len - 1] == '\n')
		--len;
	/* pat may not be NUL-terminated */
	if (wflag && !Fflag) {
		int bol = 0, eol = 0, extra;
		if (pat[0] == '^')
			bol = 1;
		if (len > 0 && pat[len - 1] == '$')
			eol = 1;
		extra = Eflag ? 2 : 4;
		pattern[patterns] = grep_malloc(len + 15 + extra);
		snprintf(pattern[patterns], len + 15 + extra,
		   "%s[[:<:]]%s%.*s%s[[:>:]]%s",
		    bol ? "^" : "",
		    Eflag ? "(" : "\\(",
		    (int)len - bol - eol, pat + bol,
		    Eflag ? ")" : "\\)",
		    eol ? "$" : "");
		len += 14 + extra;
	} else {
		pattern[patterns] = grep_malloc(len + 1);
		memcpy(pattern[patterns], pat, len);
		pattern[patterns][len] = '\0';
	}
	++patterns;
}

static void
add_patterns(char *pats)
{
	char	*nl;

	while ((nl = strchr(pats, '\n')) != NULL) {
		add_pattern(pats, nl - pats);
		pats = nl + 1;
	}
	add_pattern(pats, strlen(pats));
}

static void
read_patterns(const char *fn)
{
	FILE	*f;
	char	*line;
	size_t	 len;

	if ((f = fopen(fn, "r")) == NULL)
		err(2, "%s", fn);
	while ((line = fgetln(f, &len)) != NULL)
		add_pattern(line, *line == '\n' ? 0 : len);
	if (ferror(f))
		err(2, "%s", fn);
	fclose(f);
}

int
main(int argc, char *argv[])
{
	int		 c, lastc, prevoptind, newarg, i, needpattern;
	struct patfile	*patfile, *pf_next;
	char		*ep;
	struct stat	*finfo = 0;
	unsigned long long l;

	setlocale(LC_ALL, "");

#ifndef WITHOUT_NLS
	catalog = catopen("grep", NL_CAT_LOCALE);
#endif

	SLIST_INIT(&patfilelh);
	switch (__progname[0]) {
	case 'e':
		Eflag++;
		break;
	case 'f':
		Fflag++;
		break;
	case 'g':
		Gflag++;
		break;
	case 'z':
		Zflag++;
		switch(__progname[1]) {
		case 'e':
			Eflag++;
			break;
		case 'f':
			Fflag++;
			break;
		case 'g':
			Gflag++;
			break;
		}
		break;
	}

	lastc = '\0';
	newarg = 1;
	prevoptind = 1;
	needpattern = 1;
	while ((c = getopt_long(argc, argv, optstr,
				long_options, NULL)) != -1) {
		switch (c) {
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			if (newarg || !isdigit(lastc))
				Aflag = 0;
			else if (Aflag > LLONG_MAX / 10)
				errx(2, getstr(9));
			Aflag = Bflag = (Aflag * 10) + (c - '0');
			break;
		case 'C':
			if (optarg == NULL) {
				Aflag = Bflag = 2;
				break;
			}
			/* FALLTHROUGH */
		case 'A':
		case 'B':
			l = strtoull(optarg, &ep, 10);
			if ((errno == ERANGE) && (l == ULLONG_MAX))
				errx(2, getstr(9));
			if (c == 'A')
				Aflag = l;
			else if (c == 'B')
				Bflag = l;
			else
				Aflag = Bflag = l;
			break;
		case 'a':
			binbehave = BINFILE_TEXT;
			break;
		case 'b':
			bflag = 1;
			break;
		case 'c':
			cflag = 1;
			break;
		case 'D':
			if (strcmp(optarg, "skip") == 0)
				devbehave = DEV_SKIP;
			break;
		case 'd':
			if (strcmp("recurse", optarg) == 0) {
				Hflag++;
				dirbehave = DIR_RECURSE;
			} else if (strcmp("skip", optarg) == 0)
				dirbehave = DIR_SKIP;
			else if (strcmp("read", optarg) != 0)
				errx(2, getstr(13));
			break;
		case 'E':
			Fflag = Gflag = 0;
			Eflag++;
			break;
		case 'e':
			add_patterns(optarg);
			needpattern = 0;
			break;
		case 'F':
			Eflag = Gflag = 0;
			Fflag++;
			break;
		case 'f':
			patfile = grep_malloc(sizeof(*patfile));
			patfile->pf_file = optarg;
			SLIST_INSERT_HEAD(&patfilelh, patfile, pf_next);
			needpattern = 0;
			break;
		case 'G':
			Eflag = Fflag = 0;
			Gflag++;
			break;
		case 'H':
			Hflag++;
			break;
		case 'h':
			Hflag = 0;
			hflag = 1;
			break;
		case 'I':
			binbehave = BINFILE_SKIP;
			break;
		case 'i':
		case 'y':
			iflag = 1;
			cflags |= REG_ICASE;
			break;
		case 'J':
			Zflag = 0;
			Jflag++;
			break;
		case 'L':
			lflag = 0;
			Lflag = qflag = 1;
			break;
		case 'l':
			Lflag = 0;
			lflag = qflag = 1;
			break;
		case 'm':
			mflag++;
			mcount = strtoull(optarg, (char **)NULL, 10);
			if ((errno == ERANGE) && (mcount == ULLONG_MAX))
				err(2, getstr(13));
			break;
		case 'n':
			nflag = 1;
			break;
		case 'O':
			linkbehave = LINK_EXPLICIT;
			break;
		case 'o':
			oflag++;
			break;
		case 'P':
			linkbehave = LINK_SKIP;
			break;
		case 'q':
			qflag = 1;
			break;
		case 'S':
			linkbehave = LINK_GREP;
			break;
		case 'R':
		case 'r':
			dirbehave = DIR_RECURSE;
			Hflag++;
			break;
		case 's':
			sflag = 1;
			break;
		case 'U':
			binbehave = BINFILE_BIN;
			break;
		case 'u':
		case MMAP_OPT:
			/* default, compatibility */
			break;
		case 'V':
			printf(getstr(10));
			exit(0);
		case 'v':
			vflag = 1;
			break;
		case 'w':
			wflag = 1;
			break;
		case 'x':
			xflag = 1;
			break;
		case 'Z':
			Jflag = 0;
			Zflag++;
			break;
		case BIN_OPT:
			if (strcmp("binary", optarg) == 0)
				binbehave = BINFILE_BIN;
			else if (strcmp("without-match", optarg) == 0)
				binbehave = BINFILE_SKIP;
			else if (strcmp("text", optarg) == 0)
				binbehave = BINFILE_TEXT;
			else
				errx(2, getstr(11));
			break;
		case COLOR_OPT:
			if (optarg == NULL)
				optarg = "auto";
			if (strcmp("auto", optarg) == 0 || strcmp("always", optarg) == 0 ) {
				color = getenv("GREP_COLOR");
				if (color == NULL)
					color = "01;31";
			} else if (strcmp("never", optarg) == 0)
				color = NULL;
			else
				errx(2, getstr(14));
			break;
		case LABEL_OPT:
			label = optarg;
			break;
		case LINEBUF_OPT:
			lbflag = 1;
			break;
		case NULL_OPT:
			nullflag = 1;
			break;
		case HELP_OPT:
		default:
			usage();
		}
		lastc = c;
		newarg = optind != prevoptind;
		prevoptind = optind;
	}
	argc -= optind;
	argv += optind;

	for (patfile = SLIST_FIRST(&patfilelh); patfile != NULL;
	    patfile = pf_next) {
		pf_next = SLIST_NEXT(patfile, pf_next);
		read_patterns(patfile->pf_file);
		free(patfile);
	}

	if (argc == 0 && needpattern)
		usage();

	if (argc != 0 && needpattern) {
		add_patterns(*argv);
		--argc;
		++argv;
	}

	if (Eflag)
		cflags |= REG_EXTENDED;
	fg_pattern = grep_calloc(patterns, sizeof(*fg_pattern));
	r_pattern = grep_calloc(patterns, sizeof(*r_pattern));
	for (i = 0; i < patterns; ++i) {
		/* Check if cheating is allowed (always is for fgrep). */
		if (Fflag) {
			fgrepcomp(&fg_pattern[i], pattern[i]);
		} else {
			if (fastcomp(&fg_pattern[i], pattern[i])) {
				/* Fall back to full regex library */
				c = regcomp(&r_pattern[i], pattern[i], cflags);
				if (c != 0) {
					regerror(c, &r_pattern[i], re_error,
					    RE_ERROR_BUF);
					errx(2, "%s", re_error);
				}
			}
		}
	}

	if (lbflag)
		setlinebuf(stdout);

	if ((argc == 0 || argc == 1) && !Hflag)
		hflag = 1;

	if (argc == 0)
		exit(!procfile(NULL));

	if (dirbehave == DIR_RECURSE)
		c = grep_tree(argv);
	else
		for (c = 0; argc--; ++argv) {
			if (devbehave == DEV_SKIP) {
				finfo = grep_malloc(sizeof(struct stat));
				if (stat(*argv, finfo) == -1)
					err(2, NULL);
				if (S_ISBLK(finfo->st_mode) ||
				    S_ISCHR(finfo->st_mode) ||
				    S_ISFIFO(finfo->st_mode) ||
				    S_ISSOCK(finfo->st_mode)) {
					free(finfo);
					continue;
				} else
					free(finfo);
			}
			c+= procfile(*argv);
		}

#ifndef WITHOUT_NLS
	catclose(catalog);
#endif

	if (c) {
		if (notfound && qflag)
			exit(0);
		else if (notfound)
			exit (2);
		else
			exit (0);		
	} else if (notfound)
		exit(2);
	else
		exit(1);
}
